#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <udis86.h>
#include <sys/mman.h>
#include <errno.h>
#include "bitmapped_patricia_tree.h"
#define __USE_GNU
#include <ucontext.h>

#include "lists.h"

/* This code is only for x86_64 gcc linux. It is NOT threadsafe. */

void* slice_alloc (size_t length) {
  static intptr_t *nextslice = NULL + 1;
  static intptr_t *lastslice = NULL;

  int num = length / sizeof(intptr_t) + 1;

  if (lastslice < nextslice + num) {
    nextslice = (intptr_t *)malloc(1024*sizeof(intptr_t));
    lastslice = nextslice + 1023;
  }

  void* ret = (void*) nextslice;
  nextslice += num;
  return ret;
}

/* w00t, I am a three-star-programmer! */
void ***dereferencedPointer;
void ***lastDereferencedPointer;
void **lockedPointer;
void **lastLockedPointer;

typedef struct {
  void* (*function) (void*);
  void* argument;
  void*** dereferencedPointer;
} tree_entry;

bpt_t pointer_tree;

bpt_t bpt_assoc_and_release(bpt_t bpt, bpt_key_t key, void *value) {
  bpt_t new_bpt = bpt_assoc(bpt, key, value);
  bpt_release(bpt);
  return new_bpt;
}

void maybeAllocateMorePointers () {
  if (lastDereferencedPointer > dereferencedPointer) {
    lockedPointer = (void**) mmap(NULL, 
				  sysconf(_SC_PAGE_SIZE)*sizeof(void*),
				  PROT_NONE, MAP_ANONYMOUS | MAP_PRIVATE,
				  -1, 0);
    dereferencedPointer = (void***) mmap(NULL,
					 sysconf(_SC_PAGE_SIZE)*sizeof(void**),
					 PROT_READ | PROT_WRITE,
					 MAP_ANONYMOUS | MAP_PRIVATE, -1, 0);

    if ((dereferencedPointer == MAP_FAILED)) {
      printf("Mmap failed 2!\n");
      if ((lockedPointer == MAP_FAILED)) {
	printf("Mmap failed 1!\n");
	exit(-1);
      }
      exit(-1);
    }

    lastLockedPointer = lockedPointer + sysconf(_SC_PAGE_SIZE);
    lastDereferencedPointer = dereferencedPointer + sysconf(_SC_PAGE_SIZE);

    void** lp = lockedPointer;
    void*** dp = dereferencedPointer;
    while (lp < lastLockedPointer) {
      *dp = lp;
      lp++; dp++;
    }
  }
}

inline void initializePointerTree () {
  pointer_tree = NULL;
  dereferencedPointer = NULL + 2;
  lastDereferencedPointer = 1 + dereferencedPointer;
  maybeAllocateMorePointers ();
}

void*** lazy_alloc (void* (*calculate) (void*), void* opt) {
  maybeAllocateMorePointers ();
  void*** ret = dereferencedPointer;

  lockedPointer++;
  dereferencedPointer++;

  tree_entry *en = (tree_entry*) slice_alloc (sizeof(tree_entry));
  en->function = calculate;
  en->argument = opt;
  en->dereferencedPointer = ret;


  pointer_tree = bpt_assoc_and_release(pointer_tree, (intptr_t) *ret,
				       (void*) en);
  return ret;
}

void initializeSignalHandler();

void handle_segv(int segv, siginfo_t* siginfo, void* ucontext) {
  ucontext_t* uc = (ucontext_t*) ucontext;

  /* do disassembly at from IP */
  ud_t ud_obj;
  ud_init(&ud_obj);
  ud_set_mode(&ud_obj, 64);
  ud_set_syntax(&ud_obj, UD_SYN_ATT);
  ud_set_input_buffer(&ud_obj, (unsigned char*) uc->uc_mcontext.gregs[REG_RIP],
		      10);

  /* was disassembly successful? */
  if (!ud_disassemble(&ud_obj)) {
    printf("Disassembly fail!\n");
    exit(-1);
  }

  /* is disassembly a memory-operation? */
  struct ud_operand op0 = ud_obj.operand[0];
  struct ud_operand op1 = ud_obj.operand[1];
  struct ud_operand op;

  if (op0.type == UD_OP_MEM) {
    op = op0;
  } else if (op1.type == UD_OP_MEM) {
    op = op1;
  } else {
    printf("Instruction unknown\n");
    exit(-1);
  }

  /* find out the register - this part is clumsy as we have two sets
     of constants */

  int setreg;
  switch (op.base) {
  case UD_R_RAX:
  case UD_R_EAX:
    setreg = REG_RAX;
    break;
  case UD_R_RCX:
  case UD_R_ECX:
    setreg = REG_RCX;
    break;
  case UD_R_RDX:
  case UD_R_EDX:
    setreg = REG_RDX;
    break;
  case UD_R_RBX:
  case UD_R_EBX:
    setreg = REG_RBX;
    break;
  case UD_R_RSP:
  case UD_R_ESP:
    setreg = REG_RSP;
    break;
  case UD_R_RBP:
  case UD_R_EBP:
    setreg = REG_RBP;
    break;
  case UD_R_RSI:
  case UD_R_ESI:
    setreg = REG_RSI;
    break;
  case UD_R_RDI:
  case UD_R_EDI:
    setreg = REG_RDI;
    break;
  case UD_R_R8:
  case UD_R_R8D:
    setreg = REG_R8;
    break;
  case UD_R_R9:
  case UD_R_R9D:
    setreg = REG_R9;
    break;
  case UD_R_R10:
  case UD_R_R10D:
    setreg = REG_R10;
    break;
  case UD_R_R11:
  case UD_R_R11D:
    setreg = REG_R11;
    break;
  case UD_R_R12:
  case UD_R_R12D:
    setreg = REG_R12;
    break;
  case UD_R_R13:
  case UD_R_R13D:
    setreg = REG_R13;
    break;
  case UD_R_R14:
  case UD_R_R14D:
    setreg = REG_R14;
    break;
  case UD_R_R15:
  case UD_R_R15D:
    setreg = REG_R15;
    break;
  default:
    printf("Register not supported!\n");
    exit(-1);
  }

  intptr_t address = uc->uc_mcontext.gregs[setreg];

  if (!bpt_has_key(pointer_tree, address)) {
    printf("Address not found in Patricia tree: %d.\n", address);
    exit(-1);
  }

  tree_entry *te = (tree_entry*) bpt_get(pointer_tree, address);
  void* newAddress = (te->function)(te->argument);
  *(te->dereferencedPointer) = newAddress;

  /* set the register - as before */
  uc->uc_mcontext.gregs[setreg] = (greg_t) newAddress;
}

inline void initializeSignalHandler () {
  struct sigaction q;
  bzero(&q, sizeof(q));
  q.sa_sigaction = handle_segv;
  q.sa_flags = SA_SIGINFO | SA_NODEFER;
  sigaction(11, &q, NULL);
}

typedef struct {
  int i;
  int (*fun) (int, list);
  list initial;
} listComprehensionRecursorR;

void * listComprehensionRecursor (void* x) {
  listComprehensionRecursorR* lcr = (listComprehensionRecursorR*) x;
  int value = (lcr->fun)(lcr->i, lcr->initial);
  lcr->i++;

  list cdr = (list) lazy_alloc(listComprehensionRecursor, (void*)lcr);
  cons * ret = (cons*) slice_alloc(sizeof(cons));
  doCons(value, cdr, ret);
  return (void*) ret;
}

list listComprehension (int (*fun) (int, list)) {
  listComprehensionRecursorR* lcr =
    (listComprehensionRecursorR*) slice_alloc
    (sizeof(listComprehensionRecursorR));
  lcr->i = 0;
  lcr->fun=fun;
  lcr->initial = (list) lazy_alloc(listComprehensionRecursor, (void*)lcr);
  return lcr->initial;
}

int fib (int x, list l) {
  return x == 0 ? 0 : (x == 1 ? 1 : elt(l, x-1) + elt(l, x-2));
}

int main (void) {
  initializePointerTree ();
  initializeSignalHandler ();
  list lst = listComprehension(fib);
  int i;
  for (i = 0; i < 10; i++)
    printf("%d\n", elt(lst, i));
}
